@@ -1,9 +1,11 @@
+use forward_ref::{forward_ref_binop, forward_ref_op_assign};
 use schemars::JsonSchema;
 use serde::{de, ser, Deserialize, Deserializer, Serialize};
 use std::convert::{TryFrom, TryInto};
 use std::fmt::{self};
-use std::iter::Sum;
-use std::ops;
+use std::ops::{
+    Add, AddAssign, Div, DivAssign, Mul, MulAssign, Rem, RemAssign, Shr, ShrAssign, Sub, SubAssign,
+};
 
 use crate::errors::{DivideByZeroError, OverflowError, OverflowOperation, StdError};
 use crate::Uint128;
@@ -47,10 +49,24 @@ impl Uint64 {
         self.0
     }
 
+    /// Returns a copy of the number as big endian bytes.
+    pub const fn to_be_bytes(self) -> [u8; 8] {
+        self.0.to_be_bytes()
+    }
+
+    /// Returns a copy of the number as little endian bytes.
+    pub const fn to_le_bytes(self) -> [u8; 8] {
+        self.0.to_le_bytes()
+    }
+
     pub fn is_zero(&self) -> bool {
         self.0 == 0
     }
 
+    pub fn pow(self, exp: u32) -> Self {
+        self.0.pow(exp).into()
+    }
+
     pub fn checked_add(self, other: Self) -> Result<Self, OverflowError> {
         self.0
             .checked_add(other.0)
@@ -72,6 +88,13 @@ impl Uint64 {
             .ok_or_else(|| OverflowError::new(OverflowOperation::Mul, self, other))
     }
 
+    pub fn checked_pow(self, exp: u32) -> Result<Self, OverflowError> {
+        self.0
+            .checked_pow(exp)
+            .map(Self)
+            .ok_or_else(|| OverflowError::new(OverflowOperation::Pow, self, exp))
+    }
+
     pub fn checked_div(self, other: Self) -> Result<Self, DivideByZeroError> {
         self.0
             .checked_div(other.0)
@@ -180,11 +203,11 @@ impl From<Uint64> for u64 {
 
 impl fmt::Display for Uint64 {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
-        write!(f, "{}", self.0)
+        self.0.fmt(f)
     }
 }
 
-impl ops::Add<Uint64> for Uint64 {
+impl Add<Uint64> for Uint64 {
     type Output = Self;
 
     fn add(self, rhs: Self) -> Self {
@@ -192,7 +215,7 @@ impl ops::Add<Uint64> for Uint64 {
     }
 }
 
-impl<'a> ops::Add<&'a Uint64> for Uint64 {
+impl<'a> Add<&'a Uint64> for Uint64 {
     type Output = Self;
 
     fn add(self, rhs: &'a Uint64) -> Self {
@@ -200,7 +223,47 @@ impl<'a> ops::Add<&'a Uint64> for Uint64 {
     }
 }
 
-impl ops::Div<Uint64> for Uint64 {
+impl Sub<Uint64> for Uint64 {
+    type Output = Self;
+
+    fn sub(self, rhs: Self) -> Self {
+        Uint64(
+            self.u64()
+                .checked_sub(rhs.u64())
+                .expect("attempt to subtract with overflow"),
+        )
+    }
+}
+forward_ref_binop!(impl Sub, sub for Uint64, Uint64);
+
+impl SubAssign<Uint64> for Uint64 {
+    fn sub_assign(&mut self, rhs: Uint64) {
+        *self = *self - rhs;
+    }
+}
+forward_ref_op_assign!(impl SubAssign, sub_assign for Uint64, Uint64);
+
+impl Mul<Uint64> for Uint64 {
+    type Output = Self;
+
+    fn mul(self, rhs: Self) -> Self::Output {
+        Self(
+            self.u64()
+                .checked_mul(rhs.u64())
+                .expect("attempt to multiply with overflow"),
+        )
+    }
+}
+forward_ref_binop!(impl Mul, mul for Uint64, Uint64);
+
+impl MulAssign<Uint64> for Uint64 {
+    fn mul_assign(&mut self, rhs: Self) {
+        *self = *self * rhs;
+    }
+}
+forward_ref_op_assign!(impl MulAssign, mul_assign for Uint64, Uint64);
+
+impl Div<Uint64> for Uint64 {
     type Output = Self;
 
     fn div(self, rhs: Self) -> Self::Output {
@@ -208,7 +271,7 @@ impl ops::Div<Uint64> for Uint64 {
     }
 }
 
-impl<'a> ops::Div<&'a Uint64> for Uint64 {
+impl<'a> Div<&'a Uint64> for Uint64 {
     type Output = Self;
 
     fn div(self, rhs: &'a Uint64) -> Self::Output {
@@ -216,7 +279,27 @@ impl<'a> ops::Div<&'a Uint64> for Uint64 {
     }
 }
 
-impl ops::Shr<u32> for Uint64 {
+impl Rem for Uint64 {
+    type Output = Self;
+
+    /// # Panics
+    ///
+    /// This operation will panic if `rhs` is zero.
+    #[inline]
+    fn rem(self, rhs: Self) -> Self {
+        Self(self.0.rem(rhs.0))
+    }
+}
+forward_ref_binop!(impl Rem, rem for Uint64, Uint64);
+
+impl RemAssign<Uint64> for Uint64 {
+    fn rem_assign(&mut self, rhs: Uint64) {
+        *self = *self % rhs;
+    }
+}
+forward_ref_op_assign!(impl RemAssign, rem_assign for Uint64, Uint64);
+
+impl Shr<u32> for Uint64 {
     type Output = Self;
 
     fn shr(self, rhs: u32) -> Self::Output {
@@ -224,7 +307,7 @@ impl ops::Shr<u32> for Uint64 {
     }
 }
 
-impl<'a> ops::Shr<&'a u32> for Uint64 {
+impl<'a> Shr<&'a u32> for Uint64 {
     type Output = Self;
 
     fn shr(self, rhs: &'a u32) -> Self::Output {
@@ -232,37 +315,37 @@ impl<'a> ops::Shr<&'a u32> for Uint64 {
     }
 }
 
-impl ops::AddAssign<Uint64> for Uint64 {
+impl AddAssign<Uint64> for Uint64 {
     fn add_assign(&mut self, rhs: Uint64) {
         self.0 = self.0.checked_add(rhs.u64()).unwrap();
     }
 }
 
-impl<'a> ops::AddAssign<&'a Uint64> for Uint64 {
+impl<'a> AddAssign<&'a Uint64> for Uint64 {
     fn add_assign(&mut self, rhs: &'a Uint64) {
         self.0 = self.0.checked_add(rhs.u64()).unwrap();
     }
 }
 
-impl ops::DivAssign<Uint64> for Uint64 {
+impl DivAssign<Uint64> for Uint64 {
     fn div_assign(&mut self, rhs: Self) {
         self.0 = self.0.checked_div(rhs.u64()).unwrap();
     }
 }
 
-impl<'a> ops::DivAssign<&'a Uint64> for Uint64 {
+impl<'a> DivAssign<&'a Uint64> for Uint64 {
     fn div_assign(&mut self, rhs: &'a Uint64) {
         self.0 = self.0.checked_div(rhs.u64()).unwrap();
     }
 }
 
-impl ops::ShrAssign<u32> for Uint64 {
+impl ShrAssign<u32> for Uint64 {
     fn shr_assign(&mut self, rhs: u32) {
         self.0 = self.0.checked_shr(rhs).unwrap();
     }
 }
 
-impl<'a> ops::ShrAssign<&'a u32> for Uint64 {
+impl<'a> ShrAssign<&'a u32> for Uint64 {
     fn shr_assign(&mut self, rhs: &'a u32) {
         self.0 = self.0.checked_shr(*rhs).unwrap();
     }
@@ -345,15 +428,12 @@ impl<'de> de::Visitor<'de> for Uint64Visitor {
     }
 }
 
-impl Sum<Uint64> for Uint64 {
-    fn sum<I: Iterator<Item = Uint64>>(iter: I) -> Self {
-        iter.fold(Uint64::zero(), ops::Add::add)
-    }
-}
-
-impl<'a> Sum<&'a Uint64> for Uint64 {
-    fn sum<I: Iterator<Item = &'a Uint64>>(iter: I) -> Self {
-        iter.fold(Uint64::zero(), ops::Add::add)
+impl<A> std::iter::Sum<A> for Uint64
+where
+    Self: Add<A, Output = Self>,
+{
+    fn sum<I: Iterator<Item = A>>(iter: I) -> Self {
+        iter.fold(Self::zero(), Add::add)
     }
 }
 
@@ -405,6 +485,42 @@ mod tests {
         assert_eq!(a.to_string(), "0");
     }
 
+    #[test]
+    fn uint64_display_padding_works() {
+        let a = Uint64::from(123u64);
+        assert_eq!(format!("Embedded: {:05}", a), "Embedded: 00123");
+    }
+
+    #[test]
+    fn uint64_to_be_bytes_works() {
+        assert_eq!(Uint64::zero().to_be_bytes(), [0, 0, 0, 0, 0, 0, 0, 0]);
+        assert_eq!(
+            Uint64::MAX.to_be_bytes(),
+            [0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff]
+        );
+        assert_eq!(Uint64::new(1).to_be_bytes(), [0, 0, 0, 0, 0, 0, 0, 1]);
+        // Python: `[b for b in (63374607431768124608).to_bytes(8, "big")]`
+        assert_eq!(
+            Uint64::new(874607431768124608).to_be_bytes(),
+            [12, 35, 58, 211, 72, 116, 172, 192]
+        );
+    }
+
+    #[test]
+    fn uint64_to_le_bytes_works() {
+        assert_eq!(Uint64::zero().to_le_bytes(), [0, 0, 0, 0, 0, 0, 0, 0]);
+        assert_eq!(
+            Uint64::MAX.to_le_bytes(),
+            [0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff]
+        );
+        assert_eq!(Uint64::new(1).to_le_bytes(), [1, 0, 0, 0, 0, 0, 0, 0]);
+        // Python: `[b for b in (240282366920938463463374607431768124608).to_bytes(16, "little")]`
+        assert_eq!(
+            Uint64::new(874607431768124608).to_le_bytes(),
+            [192, 172, 116, 72, 211, 58, 35, 12]
+        );
+    }
+
     #[test]
     fn uint64_is_zero_works() {
         assert!(Uint64::zero().is_zero());
@@ -462,6 +578,84 @@ mod tests {
         assert_eq!((operand1, operand2), (a.to_string(), b.to_string()));
     }
 
+    #[test]
+    #[allow(clippy::op_ref)]
+    fn uint64_sub_works() {
+        assert_eq!(Uint64(2) - Uint64(1), Uint64(1));
+        assert_eq!(Uint64(2) - Uint64(0), Uint64(2));
+        assert_eq!(Uint64(2) - Uint64(2), Uint64(0));
+
+        // works for refs
+        let a = Uint64::new(10);
+        let b = Uint64::new(3);
+        let expected = Uint64::new(7);
+        assert_eq!(a - b, expected);
+        assert_eq!(a - &b, expected);
+        assert_eq!(&a - b, expected);
+        assert_eq!(&a - &b, expected);
+    }
+
+    #[test]
+    #[should_panic]
+    fn uint64_sub_overflow_panics() {
+        let _ = Uint64(1) - Uint64(2);
+    }
+
+    #[test]
+    fn uint64_sub_assign_works() {
+        let mut a = Uint64(14);
+        a -= Uint64(2);
+        assert_eq!(a, Uint64(12));
+
+        // works for refs
+        let mut a = Uint64::new(10);
+        let b = Uint64::new(3);
+        let expected = Uint64::new(7);
+        a -= &b;
+        assert_eq!(a, expected);
+    }
+
+    #[test]
+    #[allow(clippy::op_ref)]
+    fn uint64_mul_works() {
+        assert_eq!(Uint64::from(2u32) * Uint64::from(3u32), Uint64::from(6u32));
+        assert_eq!(Uint64::from(2u32) * Uint64::zero(), Uint64::zero());
+
+        // works for refs
+        let a = Uint64::from(11u32);
+        let b = Uint64::from(3u32);
+        let expected = Uint64::from(33u32);
+        assert_eq!(a * b, expected);
+        assert_eq!(a * &b, expected);
+        assert_eq!(&a * b, expected);
+        assert_eq!(&a * &b, expected);
+    }
+
+    #[test]
+    fn uint64_mul_assign_works() {
+        let mut a = Uint64::from(14u32);
+        a *= Uint64::from(2u32);
+        assert_eq!(a, Uint64::from(28u32));
+
+        // works for refs
+        let mut a = Uint64::from(10u32);
+        let b = Uint64::from(3u32);
+        a *= &b;
+        assert_eq!(a, Uint64::from(30u32));
+    }
+
+    #[test]
+    fn uint64_pow_works() {
+        assert_eq!(Uint64::from(2u32).pow(2), Uint64::from(4u32));
+        assert_eq!(Uint64::from(2u32).pow(10), Uint64::from(1024u32));
+    }
+
+    #[test]
+    #[should_panic]
+    fn uint64_pow_overflow_panics() {
+        Uint64::MAX.pow(2u32);
+    }
+
     #[test]
     #[should_panic]
     fn uint64_math_overflow_panics() {
@@ -475,11 +669,10 @@ mod tests {
         let base = Uint64(500);
 
         // factor 1/1
-        assert_eq!(base.multiply_ratio(1u64, 1u64), Uint64(500));
-        assert_eq!(base.multiply_ratio(3u64, 3u64), Uint64(500));
-        assert_eq!(base.multiply_ratio(654321u64, 654321u64), Uint64(500));
-        // Reactivate after https://github.com/CosmWasm/cosmwasm/issues/920
-        // assert_eq!(base.multiply_ratio(u64::MAX, u64::MAX), Uint64(500));
+        assert_eq!(base.multiply_ratio(1u64, 1u64), base);
+        assert_eq!(base.multiply_ratio(3u64, 3u64), base);
+        assert_eq!(base.multiply_ratio(654321u64, 654321u64), base);
+        assert_eq!(base.multiply_ratio(u64::MAX, u64::MAX), base);
 
         // factor 3/2
         assert_eq!(base.multiply_ratio(3u64, 2u64), Uint64(750));
@@ -533,7 +726,7 @@ mod tests {
     fn uint64_methods() {
         // checked_*
         assert!(matches!(
-            Uint64(u64::MAX).checked_add(Uint64(1)),
+            Uint64::MAX.checked_add(Uint64(1)),
             Err(OverflowError { .. })
         ));
         assert!(matches!(
@@ -541,35 +734,94 @@ mod tests {
             Err(OverflowError { .. })
         ));
         assert!(matches!(
-            Uint64(u64::MAX).checked_mul(Uint64(2)),
+            Uint64::MAX.checked_mul(Uint64(2)),
             Err(OverflowError { .. })
         ));
         assert!(matches!(
-            Uint64(u64::MAX).checked_div(Uint64(0)),
+            Uint64::MAX.checked_pow(2u32),
+            Err(OverflowError { .. })
+        ));
+        assert!(matches!(
+            Uint64::MAX.checked_div(Uint64(0)),
             Err(DivideByZeroError { .. })
         ));
         assert!(matches!(
-            Uint64(u64::MAX).checked_div_euclid(Uint64(0)),
+            Uint64::MAX.checked_div_euclid(Uint64(0)),
             Err(DivideByZeroError { .. })
         ));
         assert!(matches!(
-            Uint64(u64::MAX).checked_rem(Uint64(0)),
+            Uint64::MAX.checked_rem(Uint64(0)),
             Err(DivideByZeroError { .. })
         ));
 
         // saturating_*
-        assert_eq!(Uint64(u64::MAX).saturating_add(Uint64(1)), Uint64(u64::MAX));
+        assert_eq!(Uint64::MAX.saturating_add(Uint64(1)), Uint64::MAX);
         assert_eq!(Uint64(0).saturating_sub(Uint64(1)), Uint64(0));
-        assert_eq!(Uint64(u64::MAX).saturating_mul(Uint64(2)), Uint64(u64::MAX));
-        assert_eq!(Uint64(u64::MAX).saturating_pow(2), Uint64(u64::MAX));
+        assert_eq!(Uint64::MAX.saturating_mul(Uint64(2)), Uint64::MAX);
+        assert_eq!(Uint64::MAX.saturating_pow(2), Uint64::MAX);
 
         // wrapping_*
-        assert_eq!(Uint64(u64::MAX).wrapping_add(Uint64(1)), Uint64(0));
-        assert_eq!(Uint64(0).wrapping_sub(Uint64(1)), Uint64(u64::MAX));
+        assert_eq!(Uint64::MAX.wrapping_add(Uint64(1)), Uint64(0));
+        assert_eq!(Uint64(0).wrapping_sub(Uint64(1)), Uint64::MAX);
+        assert_eq!(Uint64::MAX.wrapping_mul(Uint64(2)), Uint64(u64::MAX - 1));
+        assert_eq!(Uint64::MAX.wrapping_pow(2), Uint64(1));
+    }
+
+    #[test]
+    #[allow(clippy::op_ref)]
+    fn uint64_implements_rem() {
+        let a = Uint64::new(10);
+        assert_eq!(a % Uint64::new(10), Uint64::zero());
+        assert_eq!(a % Uint64::new(2), Uint64::zero());
+        assert_eq!(a % Uint64::new(1), Uint64::zero());
+        assert_eq!(a % Uint64::new(3), Uint64::new(1));
+        assert_eq!(a % Uint64::new(4), Uint64::new(2));
+
+        // works for refs
+        let a = Uint64::new(10);
+        let b = Uint64::new(3);
+        let expected = Uint64::new(1);
+        assert_eq!(a % b, expected);
+        assert_eq!(a % &b, expected);
+        assert_eq!(&a % b, expected);
+        assert_eq!(&a % &b, expected);
+    }
+
+    #[test]
+    #[should_panic(expected = "divisor of zero")]
+    fn uint64_rem_panics_for_zero() {
+        let _ = Uint64::new(10) % Uint64::zero();
+    }
+
+    #[test]
+    #[allow(clippy::op_ref)]
+    fn uint64_rem_works() {
         assert_eq!(
-            Uint64(u64::MAX).wrapping_mul(Uint64(2)),
-            Uint64(u64::MAX - 1)
+            Uint64::from(12u32) % Uint64::from(10u32),
+            Uint64::from(2u32)
         );
-        assert_eq!(Uint64(u64::MAX).wrapping_pow(2), Uint64(1));
+        assert_eq!(Uint64::from(50u32) % Uint64::from(5u32), Uint64::zero());
+
+        // works for refs
+        let a = Uint64::from(42u32);
+        let b = Uint64::from(5u32);
+        let expected = Uint64::from(2u32);
+        assert_eq!(a % b, expected);
+        assert_eq!(a % &b, expected);
+        assert_eq!(&a % b, expected);
+        assert_eq!(&a % &b, expected);
+    }
+
+    #[test]
+    fn uint64_rem_assign_works() {
+        let mut a = Uint64::from(30u32);
+        a %= Uint64::from(4u32);
+        assert_eq!(a, Uint64::from(2u32));
+
+        // works for refs
+        let mut a = Uint64::from(25u32);
+        let b = Uint64::from(6u32);
+        a %= &b;
+        assert_eq!(a, Uint64::from(1u32));
     }
 }
