@@ -9,7 +9,11 @@ use crate::coins::Coin;
 use crate::deps::OwnedDeps;
 use crate::errors::{RecoverPubkeyError, StdError, StdResult, SystemError, VerificationError};
 #[cfg(feature = "stargate")]
-use crate::ibc::{IbcChannel, IbcEndpoint, IbcOrder, IbcPacket, IbcTimeoutBlock};
+use crate::ibc::{
+    IbcAcknowledgement, IbcChannel, IbcChannelCloseMsg, IbcChannelConnectMsg, IbcChannelOpenMsg,
+    IbcEndpoint, IbcOrder, IbcPacket, IbcPacketAckMsg, IbcPacketReceiveMsg, IbcPacketTimeoutMsg,
+    IbcTimeoutBlock,
+};
 use crate::query::{
     AllBalanceResponse, BalanceResponse, BankQuery, CustomQuery, QueryRequest, WasmQuery,
 };
@@ -24,6 +28,7 @@ use crate::storage::MemoryStorage;
 use crate::timestamp::Timestamp;
 use crate::traits::{Api, Querier, QuerierResult};
 use crate::types::{BlockInfo, ContractInfo, Env, MessageInfo};
+use crate::Attribute;
 
 pub const MOCK_CONTRACT_ADDR: &str = "cosmos2contract";
 
@@ -55,6 +60,14 @@ pub fn mock_dependencies_with_balances(
 // We can later make simplifications here if needed
 pub type MockStorage = MemoryStorage;
 
+/// Length of canonical addresses created with this API. Contracts should not make any assumtions
+/// what this value is.
+/// The value here must be restorable with `SHUFFLES_ENCODE` + `SHUFFLES_DECODE` in-shuffles.
+const CANONICAL_LENGTH: usize = 54;
+
+const SHUFFLES_ENCODE: usize = 18;
+const SHUFFLES_DECODE: usize = 2;
+
 // MockPrecompiles zero pads all human addresses to make them fit the canonical_length
 // it trims off zeros for the reverse operation.
 // not really smart, but allows us to see a difference (and consistent length for canonical adddresses)
@@ -62,13 +75,13 @@ pub type MockStorage = MemoryStorage;
 pub struct MockApi {
     /// Length of canonical addresses created with this API. Contracts should not make any assumtions
     /// what this value is.
-    pub canonical_length: usize,
+    canonical_length: usize,
 }
 
 impl Default for MockApi {
     fn default() -> Self {
         MockApi {
-            canonical_length: 24,
+            canonical_length: CANONICAL_LENGTH,
         }
     }
 }
@@ -100,7 +113,7 @@ impl Api for MockApi {
         // the most obvious structure (https://github.com/CosmWasm/cosmwasm/issues/552)
         let rotate_by = digit_sum(&out) % self.canonical_length;
         out.rotate_left(rotate_by);
-        for _ in 0..18 {
+        for _ in 0..SHUFFLES_ENCODE {
             out = riffle_shuffle(&out);
         }
         Ok(out.into())
@@ -115,7 +128,7 @@ impl Api for MockApi {
 
         let mut tmp: Vec<u8> = canonical.clone().into();
         // Shuffle two more times which restored the original value (24 elements are back to original after 20 rounds)
-        for _ in 0..2 {
+        for _ in 0..SHUFFLES_DECODE {
             tmp = riffle_shuffle(&tmp);
         }
         // Rotate back
@@ -208,9 +221,9 @@ pub fn mock_info(sender: &str, funds: &[Coin]) -> MessageInfo {
     }
 }
 
-#[cfg(feature = "stargate")]
 /// Creates an IbcChannel for testing. You set a few key parameters for handshaking,
 /// If you want to set more, use this as a default and mutate other fields
+#[cfg(feature = "stargate")]
 pub fn mock_ibc_channel(my_channel_id: &str, order: IbcOrder, version: &str) -> IbcChannel {
     IbcChannel {
         endpoint: IbcEndpoint {
@@ -223,16 +236,78 @@ pub fn mock_ibc_channel(my_channel_id: &str, order: IbcOrder, version: &str) ->
         },
         order,
         version: version.to_string(),
-        counterparty_version: Some(version.to_string()),
         connection_id: "connection-2".to_string(),
     }
 }
 
+/// Creates a IbcChannelOpenMsg::OpenInit for testing ibc_channel_open.
 #[cfg(feature = "stargate")]
-/// Creates a IbcPacket for testing ibc_packet_receive. You set a few key parameters that are
+pub fn mock_ibc_channel_open_init(
+    my_channel_id: &str,
+    order: IbcOrder,
+    version: &str,
+) -> IbcChannelOpenMsg {
+    IbcChannelOpenMsg::new_init(mock_ibc_channel(my_channel_id, order, version))
+}
+
+/// Creates a IbcChannelOpenMsg::OpenTry for testing ibc_channel_open.
+#[cfg(feature = "stargate")]
+pub fn mock_ibc_channel_open_try(
+    my_channel_id: &str,
+    order: IbcOrder,
+    version: &str,
+) -> IbcChannelOpenMsg {
+    IbcChannelOpenMsg::new_try(mock_ibc_channel(my_channel_id, order, version), version)
+}
+
+/// Creates a IbcChannelConnectMsg::ConnectAck for testing ibc_channel_connect.
+#[cfg(feature = "stargate")]
+pub fn mock_ibc_channel_connect_ack(
+    my_channel_id: &str,
+    order: IbcOrder,
+    version: &str,
+) -> IbcChannelConnectMsg {
+    IbcChannelConnectMsg::new_ack(mock_ibc_channel(my_channel_id, order, version), version)
+}
+
+/// Creates a IbcChannelConnectMsg::ConnectConfirm for testing ibc_channel_connect.
+#[cfg(feature = "stargate")]
+pub fn mock_ibc_channel_connect_confirm(
+    my_channel_id: &str,
+    order: IbcOrder,
+    version: &str,
+) -> IbcChannelConnectMsg {
+    IbcChannelConnectMsg::new_confirm(mock_ibc_channel(my_channel_id, order, version))
+}
+
+/// Creates a IbcChannelCloseMsg::CloseInit for testing ibc_channel_close.
+#[cfg(feature = "stargate")]
+pub fn mock_ibc_channel_close_init(
+    my_channel_id: &str,
+    order: IbcOrder,
+    version: &str,
+) -> IbcChannelCloseMsg {
+    IbcChannelCloseMsg::new_init(mock_ibc_channel(my_channel_id, order, version))
+}
+
+/// Creates a IbcChannelCloseMsg::CloseConfirm for testing ibc_channel_close.
+#[cfg(feature = "stargate")]
+pub fn mock_ibc_channel_close_confirm(
+    my_channel_id: &str,
+    order: IbcOrder,
+    version: &str,
+) -> IbcChannelCloseMsg {
+    IbcChannelCloseMsg::new_confirm(mock_ibc_channel(my_channel_id, order, version))
+}
+
+/// Creates a IbcPacketReceiveMsg for testing ibc_packet_receive. You set a few key parameters that are
 /// often parsed. If you want to set more, use this as a default and mutate other fields
-pub fn mock_ibc_packet_recv<T: Serialize>(my_channel_id: &str, data: &T) -> StdResult<IbcPacket> {
-    Ok(IbcPacket {
+#[cfg(feature = "stargate")]
+pub fn mock_ibc_packet_recv(
+    my_channel_id: &str,
+    data: &impl Serialize,
+) -> StdResult<IbcPacketReceiveMsg> {
+    Ok(IbcPacketReceiveMsg::new(IbcPacket {
         data: to_binary(data)?,
         src: IbcEndpoint {
             port_id: "their-port".to_string(),
@@ -248,14 +323,14 @@ pub fn mock_ibc_packet_recv<T: Serialize>(my_channel_id: &str, data: &T) -> StdR
             height: 12345678,
         }
         .into(),
-    })
+    }))
 }
 
-#[cfg(feature = "stargate")]
 /// Creates a IbcPacket for testing ibc_packet_{ack,timeout}. You set a few key parameters that are
 /// often parsed. If you want to set more, use this as a default and mutate other fields.
-/// The difference between mock_ibc_packet_recv is if `my_channel_id` is src or dest.
-pub fn mock_ibc_packet_ack<T: Serialize>(my_channel_id: &str, data: &T) -> StdResult<IbcPacket> {
+/// The difference from mock_ibc_packet_recv is if `my_channel_id` is src or dest.
+#[cfg(feature = "stargate")]
+fn mock_ibc_packet(my_channel_id: &str, data: &impl Serialize) -> StdResult<IbcPacket> {
     Ok(IbcPacket {
         data: to_binary(data)?,
         src: IbcEndpoint {
@@ -275,6 +350,31 @@ pub fn mock_ibc_packet_ack<T: Serialize>(my_channel_id: &str, data: &T) -> StdRe
     })
 }
 
+/// Creates a IbcPacketAckMsg for testing ibc_packet_ack. You set a few key parameters that are
+/// often parsed. If you want to set more, use this as a default and mutate other fields.
+/// The difference from mock_ibc_packet_recv is if `my_channel_id` is src or dest.
+#[cfg(feature = "stargate")]
+pub fn mock_ibc_packet_ack(
+    my_channel_id: &str,
+    data: &impl Serialize,
+    ack: IbcAcknowledgement,
+) -> StdResult<IbcPacketAckMsg> {
+    let packet = mock_ibc_packet(my_channel_id, data)?;
+
+    Ok(IbcPacketAckMsg::new(ack, packet))
+}
+
+/// Creates a IbcPacketTimeoutMsg for testing ibc_packet_timeout. You set a few key parameters that are
+/// often parsed. If you want to set more, use this as a default and mutate other fields.
+/// The difference from mock_ibc_packet_recv is if `my_channel_id` is src or dest./
+#[cfg(feature = "stargate")]
+pub fn mock_ibc_packet_timeout(
+    my_channel_id: &str,
+    data: &impl Serialize,
+) -> StdResult<IbcPacketTimeoutMsg> {
+    mock_ibc_packet(my_channel_id, data).map(IbcPacketTimeoutMsg::new)
+}
+
 /// The same type as cosmwasm-std's QuerierResult, but easier to reuse in
 /// cosmwasm-vm. It might diverge from QuerierResult at some point.
 pub type MockQuerierCustomHandlerResult = SystemResult<ContractResult<Binary>>;
@@ -311,9 +411,9 @@ impl<C: DeserializeOwned> MockQuerier<C> {
     }
 
     // set a new balance for the given address and return the old balance
-    pub fn update_balance<U: Into<String>>(
+    pub fn update_balance(
         &mut self,
-        addr: U,
+        addr: impl Into<String>,
         balance: Vec<Coin>,
     ) -> Option<Vec<Coin>> {
         self.bank.balances.insert(addr.into(), balance)
@@ -508,6 +608,49 @@ impl StakingQuerier {
 ///
 /// https://en.wikipedia.org/wiki/Riffle_shuffle_permutation#Perfect_shuffles
 /// https://en.wikipedia.org/wiki/In_shuffle
+///
+/// The number of shuffles required to restore the original order are listed in
+/// https://oeis.org/A002326, e.g.:
+///
+/// ```ignore
+/// 2: 2
+/// 4: 4
+/// 6: 3
+/// 8: 6
+/// 10: 10
+/// 12: 12
+/// 14: 4
+/// 16: 8
+/// 18: 18
+/// 20: 6
+/// 22: 11
+/// 24: 20
+/// 26: 18
+/// 28: 28
+/// 30: 5
+/// 32: 10
+/// 34: 12
+/// 36: 36
+/// 38: 12
+/// 40: 20
+/// 42: 14
+/// 44: 12
+/// 46: 23
+/// 48: 21
+/// 50: 8
+/// 52: 52
+/// 54: 20
+/// 56: 18
+/// 58: 58
+/// 60: 60
+/// 62: 6
+/// 64: 12
+/// 66: 66
+/// 68: 22
+/// 70: 35
+/// 72: 9
+/// 74: 20
+/// ```
 pub fn riffle_shuffle<T: Clone>(input: &[T]) -> Vec<T> {
     assert!(
         input.len() % 2 == 0,
@@ -527,6 +670,15 @@ pub fn digit_sum(input: &[u8]) -> usize {
     input.iter().fold(0, |sum, val| sum + (*val as usize))
 }
 
+/// Only for test code. This bypasses assertions in new, allowing us to create _*
+/// Attributes to simulate responses from the blockchain
+pub fn mock_wasmd_attr(key: impl Into<String>, value: impl Into<String>) -> Attribute {
+    Attribute {
+        key: key.into(),
+        value: value.into(),
+    }
+}
+
 #[cfg(test)]
 mod tests {
     use super::*;
@@ -582,7 +734,8 @@ mod tests {
     #[should_panic(expected = "address too long")]
     fn addr_canonicalize_max_input_length() {
         let api = MockApi::default();
-        let human = String::from("some-extremely-long-address-not-supported-by-this-api");
+        let human =
+            String::from("some-extremely-long-address-not-supported-by-this-api-longer-than-54");
         let _ = api.addr_canonicalize(&human).unwrap();
     }
 
@@ -594,7 +747,7 @@ mod tests {
         api.addr_canonicalize(&input).unwrap();
 
         let input = "foobar456";
-        api.addr_canonicalize(&input).unwrap();
+        api.addr_canonicalize(input).unwrap();
     }
 
     #[test]
@@ -802,8 +955,8 @@ mod tests {
         let signature = hex::decode(ED25519_SIG_HEX).unwrap();
         let public_key: Vec<u8> = vec![0u8; 0];
 
-        let msgs: Vec<&[u8]> = vec![&msg.as_slice()];
-        let signatures: Vec<&[u8]> = vec![&signature.as_slice()];
+        let msgs: Vec<&[u8]> = vec![msg.as_slice()];
+        let signatures: Vec<&[u8]> = vec![signature.as_slice()];
         let public_keys: Vec<&[u8]> = vec![&public_key];
 
         let res = api.ed25519_batch_verify(&msgs, &signatures, &public_keys);
@@ -959,9 +1112,9 @@ mod tests {
 
     #[cfg(feature = "staking")]
     // gets delegators from query or panic
-    fn get_all_delegators<U: Into<String>>(
+    fn get_all_delegators(
         staking: &StakingQuerier,
-        delegator: U,
+        delegator: impl Into<String>,
     ) -> Vec<Delegation> {
         let raw = staking
             .query(&StakingQuery::AllDelegations {
@@ -975,10 +1128,10 @@ mod tests {
 
     #[cfg(feature = "staking")]
     // gets full delegators from query or panic
-    fn get_delegator<U: Into<String>, V: Into<String>>(
+    fn get_delegator(
         staking: &StakingQuerier,
-        delegator: U,
-        validator: V,
+        delegator: impl Into<String>,
+        validator: impl Into<String>,
     ) -> Option<FullDelegation> {
         let raw = staking
             .query(&StakingQuery::Delegation {
